


template<typename T>
int Image<T>::id = 0;

template<typename T>
bool Image<T>::load(const char *name, bool grey)
{
#ifdef _OPENCV_
	if (depth != 1) { ERROR("Cannot load in image with depth different from 1 byte"); return false; }
	IplImage* cvImg = cvLoadImage(name, CV_LOAD_IMAGE_ANYDEPTH | CV_LOAD_IMAGE_ANYCOLOR);
	if (cvImg == NULL) { ERROR("Failed to load image " << name); return false; }
	if (cvImg->width != w || cvImg->height != h || chan != cvImg->nChannels || data_ == NULL)
	{
		destroy();
		create(cvImg->width, cvImg->height, (grey ? 1 : cvImg->nChannels));
	}
	// copy data_ from OpeCV IplImage to IPP image
	uchar *data8u = (uchar*)data_;
	if (grey && cvImg->nChannels == 3)
		for(int i = 0; i < cvImg->height; i++)
		{
			uchar *lineCV = (uchar*)(cvImg->imageData+(i*cvImg->widthStep));
			uchar *line = data8u+(i*step);
			for(int j = 0; j < step; j++) line[j] = 0.299*lineCV[3*j+2]+0.587*lineCV[3*j+1]+0.114*lineCV[3*j+0];
		}
	else
		for(int i = 0; i < h; i++)
		{
			uchar *lineCV = (uchar*)(cvImg->imageData+(i*cvImg->widthStep));
			uchar *line = data8u+(i*step);
			for(int j = 0; j < step; j++) line[j] = lineCV[j];
		}
	cvReleaseImage(&cvImg); // memory release with OpenCV
	return true;
#else
	// only read ppm and bmp

#endif
}


template<typename T>
bool Image<T>::save_show(const char *name, bool save)
{
#ifdef _OPENCV_
	// create open cv image
	CvSize sizeCv; sizeCv.width = w; sizeCv.height = h; 
	IplImage *cvImg = cvCreateImage(sizeCv, IPL_DEPTH_8U, chan); // create a cv image
	if (cvImg == NULL) { ERROR("CV image creation failed"); return false; }

	// convert to 8u
	Image<uchar> *img8u; 
	if (depth == 1) img8u = (Image<uchar>*)this; else
	{
		img8u = new Image<uchar>(w, h, chan);
		double vmin = HUGE, vmax = -HUGE;
		int tmp;
		for(int i = 0; i < w*h*chan; i++) 
			{ if (data_[i] < vmin) vmin = data_[i]; if (data_[i] > vmax) vmax = data_[i]; }
		for(int i = 0; i < w*h*chan; i++) 
		{
			tmp = (int)(((data_[i]-vmin)*255)/(vmax-vmin));
			if (tmp < 0) tmp = 0; if (tmp > 255) tmp = 255;
			img8u->data_[i] = (uchar)tmp;
		}
	}

	// swap data_
	char *sve = cvImg->imageData, *sveo = cvImg->imageDataOrigin;
	int wstep = cvImg->widthStep;
	
	cvImg->imageData = cvImg->imageDataOrigin = (char*)(img8u->data_);
	cvImg->widthStep = img8u->step;

	// do it
	if (save)
	{
		if (strstr(name, ".ppm") == (name+strlen(name)-4))
		{ // on fait la sauvegarde en ppm  la main car cvSaveImage enregistre en ppm ascii par dfaut ...)
			FILE *file = fopen(name, "wb"); ASSERT(file);
			fprintf(file, "P%c\n%d %d\n%d\n", (img8u->chan == 3 ? '6' : '5'), img8u->w, img8u->h, 255);
			int res = fwrite(img8u->data_, 1, img8u->w*img8u->h*img8u->chan, file);
			ASSERTC(res,==,img8u->w*img8u->h*img8u->chan);
			fclose(file);	
		} else
		cvSaveImage(name, cvImg); // save image to disk
	} else
	{
		cvShowImage(name, cvImg); // displays cv image
		cvWaitKey(0); // wait for a key (==0) or wait milliseconds
	}
	
	// reswap data_
	cvImg->imageData = sve, cvImg->imageDataOrigin = sveo;
	cvImg->widthStep = wstep;
	
	// release
	cvReleaseImage(&cvImg); // free memory cv image and also tmp
	if (depth != 1) delete img8u;
	return true;
#else
	// only save ppm

#endif
}


template<typename T>
bool Image<T>::save(const char *name)
{
	return save_show(name, true);
}

template<typename T>
bool Image<T>::show(const char *name)
{
	return save_show(name, false);
}

template<typename T>
void Image<T>::show()
{
#ifdef _OPENCV_
	char name[14] = "imgwin";
	for(int i = 6, tmp = id; i < 13; i++) { name[i] = tmp % 10; tmp /= 10; }
	openWindow(name); 
	show(name);
	closeWindow(name); 
#endif
}

template<typename T>
void Image<T>::openWindow(const char *name)
{
#ifdef _OPENCV_
	cvNamedWindow(name, 1); // create a window
#endif
}

template<typename T>
void Image<T>::closeWindow(const char *name)
{
#ifdef _OPENCV_
	cvDestroyWindow(name); // close cv window and free memory
#endif
}


