#include "snake-edge-mp.h"

namespace snakes {


EdgeMP::EdgeMP(real minDist, real maxDist, real stopThres, real timeStep)
{
	setProperties(minDist, maxDist, stopThres, timeStep);
	sParticles = 0;
	localProps_[0] = NULL;
	localProps_[1] = NULL;
}

void EdgeMP::setProperties(real minDist, real maxDist, real stopThres, real timeStep)
{
	this->minDist2 = sqr(min(minDist,maxDist));
	this->maxDist2 = sqr(max(minDist,maxDist));
	this->stopThres = stopThres;
	this->timeStep = timeStep;
}


void EdgeMP::initialize(cyril::Image<uchar> &img, real center[nLocalProps], real radius, int propagation, real globalProps[nGlobalProps])
{
int gradw = 1;
int gradw2 = 1;

	this->propagation = sign(propagation);
	nParticles = (int)(2*M_PI*radius / ((sqrt(minDist2)+sqrt(maxDist2))/2));
	if (nParticles > sParticles) resizeParticles(nParticles);
	
	real dtheta = 2*M_PI/nParticles, theta = 0;
	for(int i = 0; i < nParticles; i++, theta += dtheta)
	{
		for(int j = 0; j < nLocalProps; j++)
			localProps[i][j] = center[j];
		localProps[i][x] = center[x] + radius*cos(theta);// + rand()%8-4;
		localProps[i][y] = center[y] + radius*sin(theta);// + rand()%8-4;
		
		if (localProps[i][x] < (gradw+gradw2))        localProps[i][x] = (gradw+gradw2);
		if (localProps[i][x] >= img.w-(gradw+gradw2)) localProps[i][x] = img.w-(gradw+gradw2)-.01;
		if (localProps[i][y] < (gradw+gradw2))        localProps[i][y] = (gradw+gradw2);
		if (localProps[i][y] >= img.h-(gradw+gradw2)) localProps[i][y] = img.h-(gradw+gradw2)-.01;
		
	}
	
	INFO(nParticles << " particles");
	
	for(int j = 0; j < nGlobalProps; j++)
		this->globalProps[j] = globalProps[j];
}

void EdgeMP::saveResult(cyril::Image<uchar> &img, cyril::Image<uchar> &img_disp, char *prefix, int num, char *suffix)
{
		for(int i = 0; i < img.h; i++) 
		for(int j = 0; j < img.w; j++)
		for(int k = 0; k < 3; k++)
			img_disp.data(i,j,k)=img.data(i,j);
		
		for(int i = 0; i < nParticles; i++)
		{
			int x = (int)localProps[i][EdgeMP::x], y = (int)localProps[i][EdgeMP::y];
			uchar col[3] = {0,0,255};
			for(int j = -1; j <= 1; j++) for(int k = 0; k < 3; k++) 
				{ img_disp.data(y,x+j,k) = col[k]; img_disp.data(y+j,x,k) = col[k]; }
		}
		
		std::ostringstream oss;
		oss << prefix << (100000+num) << suffix;
		img_disp.save(oss.str().c_str());
		img_disp.show("pouet");
}

int EdgeMP::converge(cyril::Image<uchar> &img)
{
	cyril::Image<char>::openWindow("pouet");

	int nsteps = 0;
	cyril::Image<uchar> img_disp;
	img_disp.create(img.w, img.h, 3);
	img.save("grey.png");
	
	real max_evol;
	do {
		if (nsteps % 20 == 0)
			saveResult(img, img_disp, "result", nsteps/30, "a.jpg");
		reorganize();
//		saveResult(img, img_disp, "result", nsteps, "b.png");
		
		nsteps++;
		
		max_evol = update(img);
		
		Particle *tmplp = localProps;
		localProps = localProps_new;
		localProps_new = tmplp;
	
	} while (max_evol > stopThres);
	
	saveResult(img, img_disp, "result", nsteps, "a.png");
	
	return nsteps;
}

inline real gradN2(cyril::Image<uchar> &img, int y, int x, int width)
{
return
	sqr((img.data(y, x+width) - img.data(y, x-width)) / (2.*width)) + 
	sqr((img.data(y+width, x) - img.data(y-width, x)) / (2.*width));
}
inline real gradN2N(cyril::Image<uchar> &img, int y, int x, int width, real xn, real yn)
{
//return
//if (xn < 0.707) xn = 0;
//if (yn < 0.707) yn = 0;
	real gx = (img.data(y, x+width) - img.data(y, x-width)) / (2.);
	real gy = (img.data(y+width, x) - img.data(y-width, x)) / (2.);
	real g2 = gx*gx + gy*gy;
//	return g2;
	real g2cos2 = sqr(xn*gx + yn*gy);
//	return g2cos2;
	if (g2 < 0.001) return g2;
	real cos2 = g2cos2 / g2;
	return g2*cos2*cos2*cos2;
//	DEBUG(gx << "," << gy <<  "," <<g2<<","<<g2cos2<<","<<cos2<<","<<res);
/*	real res =
	sqr(xn * (img.data(y, x+width) - img.data(y, x-width)) / (2.) + 
	    yn * (img.data(y+width, x) - img.data(y-width, x)) / (2.));
*/
//	std::cout << res << " ";
//	return res;
}

real EdgeMP::update(cyril::Image<uchar> &img)
{
int gradw = 1;
int gradw2 = 1;
	// local properties
	real max_evol = -1;
	real min_phi = -1;
	for(int i = 0; i < nParticles; i++)
	{	
		real x = localProps[i][EdgeMP::x], y = localProps[i][EdgeMP::y];
//if ((int)y == 53) BREAK;
		real xp = localProps[(i+1)%nParticles][EdgeMP::x], yp = localProps[(i+1)%nParticles][EdgeMP::y];
		real xm = localProps[(i-1+nParticles)%nParticles][EdgeMP::x], ym = localProps[(i-1+nParticles)%nParticles][EdgeMP::y];
		
		real dx_2 = (xp-xm), dy_2 = (yp-ym);
		real ddx = xp-2*x+xm, ddy = yp-2*y+ym;
		real norm_2 = sqrt(dx_2*dx_2+dy_2*dy_2);
		real kappa = 4/(norm_2*norm_2*norm_2) * (ddx*dy_2-ddy*dx_2);
		real xn = dy_2/norm_2, yn = -dx_2/norm_2; // N
		//xn=1, yn=1;
		
		real coef_phi = 1, coef_grad = 1./2000;
		real phi = coef_phi/ (1 + coef_grad*gradN2N(img, (int)y, (int)x, gradw, xn, yn));
		real phi_xp = coef_phi / (1 + coef_grad*gradN2N(img, (int)y, (int)(x+gradw2), gradw, xn, yn));
		real phi_xm = coef_phi/ (1 + coef_grad*gradN2N(img, (int)y, (int)(x-gradw2), gradw, xn, yn));
		real phi_yp = coef_phi/ (1 + coef_grad*gradN2N(img, (int)(y+gradw2), (int)x, gradw, xn, yn));
		real phi_ym = coef_phi/ (1 + coef_grad*gradN2N(img, (int)(y-gradw2), (int)x, gradw, xn, yn));
		
		xn = dy_2/norm_2, yn = -dx_2/norm_2; // N
		real dphin = (yn*(phi_xp-phi_xm) - xn*(phi_yp-phi_ym)) / 2.;
//		real dphin = 10*(yn*(phi_xp-phi) - xn*(phi_yp-phi));

		real ct = 40.*phi*kappa - 800.*dphin + 1.*kappa + propagation*(0.1 + 0.2*phi);
//		real ct = 40.*phi*kappa - 400.*dphin + propagation*(0.0 + 0.4*phi);
		ct *= timeStep;
//		if (ct < -0.8) ct = -0.8;
//		if (ct > 0.8) ct = 0.8;
		real ct_x = ct * xn, ct_y = ct * yn;
				
		localProps_new[i][EdgeMP::x] = x + ct_x;
		localProps_new[i][EdgeMP::y] = y + ct_y;
		
		if (localProps_new[i][EdgeMP::x] < (gradw+gradw2))        localProps_new[i][EdgeMP::x] = (gradw+gradw2);
		if (localProps_new[i][EdgeMP::x] >= img.w-(gradw+gradw2)) localProps_new[i][EdgeMP::x] = img.w-(gradw+gradw2)-.01;
		if (localProps_new[i][EdgeMP::y] < (gradw+gradw2))        localProps_new[i][EdgeMP::y] = (gradw+gradw2);
		if (localProps_new[i][EdgeMP::y] >= img.h-(gradw+gradw2)) localProps_new[i][EdgeMP::y] = img.h-(gradw+gradw2)-.01;

		//if (sqr(ct*timeStep) > 10*maxDist2) BREAK;
		real evol = max(abs(x-localProps_new[i][EdgeMP::x]), abs(y-localProps_new[i][EdgeMP::y])) / timeStep;
		if (max_evol < -0.5 || evol > max_evol) max_evol = evol;
		if (min_phi < -0.5 || phi < min_phi) min_phi = phi;
	}
	// global properties
	
	// end
	INFO("max evol " << max_evol << ", min phi " << min_phi);
	return max_evol;
}


void EdgeMP::reorganize()
{
//	for(int i = 0; i < nParticles; i++) std::cout << i << " " << localProps[i][x] << " " << localProps[i][y] << " ; ";
//	std::cout << std::endl;


	for(int i = 0; i < nParticles; i++)
	{
		int ip = (i+1)%nParticles;
		real dist2 = sqr(localProps[ip][x]-localProps[i][x]) + sqr(localProps[ip][y]-localProps[i][y]);
		if (dist2 < minDist2)
		{
			INFO("remove point " << i << " (dist2 " << dist2 << ")");
			for(int j = 0; j < nLocalProps; j++)
				localProps[i][j] = (localProps[ip][j] + localProps[i][j]) / 2;
			nParticles--;
			for(int ii = ip; ii < nParticles; ii++)
				for(int j = 0; j < nLocalProps; j++)
					localProps[ii][j] = localProps[ii+1][j];
			i--;
//	for(int i = 0; i < nParticles; i++) std::cout << i << " " << localProps[i][x] << " " << localProps[i][y] << " ; ";
//	std::cout << std::endl;
		} else
		if (dist2 > maxDist2)
		{
			INFO("add point " << i << " (dist2 " << dist2 << ")");
			if (nParticles+1 > sParticles) resizeParticles();
			if (i != nParticles-1) 
			for(int ii = nParticles-1; ii >= ip; ii--)
				for(int j = 0; j < nLocalProps; j++)
					localProps[ii+1][j] = localProps[ii][j];
			nParticles++;
			int ipp = (i+2)%nParticles;
			for(int j = 0; j < nLocalProps; j++)
				localProps[i+1][j] = (localProps[i][j] + localProps[ipp][j]) / 2;
			i--;
//	for(int i = 0; i < nParticles; i++) std::cout << i << " " << localProps[i][x] << " " << localProps[i][y] << " ; ";
//	std::cout << std::endl;
		}
	}

}


void EdgeMP::resizeParticles(int nPart)
{ // resize
	int newSParticles = (nParticles > 0 ? nParticles * 2 : (sParticles > 0 ? sParticles * 2 : 20));
	Particle *new_particles[2];
	for(int k = 0; k < 2; k++)
		new_particles[k] = new Particle[newSParticles];
	
	for(int ii = 0; ii < sParticles; ii++)
		for(int j = 0; j < nLocalProps; j++)
			for(int k = 0; k < 2; k++)
				new_particles[k][ii][j] = localProps_[k][ii][j];
	if (localProps == localProps_[0] || localProps_[0] == NULL) 
		{ localProps = new_particles[0]; localProps_new = new_particles[1]; } else  
		{ localProps = new_particles[1]; localProps_new = new_particles[0]; }
	for(int k = 0; k < 2; k++)
	{
		delete[] localProps_[k];
		localProps_[k] = new_particles[k];
	}
	sParticles = newSParticles;
}


} // snakes







