#ifndef _IMPLEMENTATIONS_HPP_
#define _IMPLEMENTATIONS_HPP_

#include "cyril-image.hpp"

namespace snakes {

class EdgeMP
{
	// config marker particles
	real minDist2, maxDist2;
	real stopThres;
	real timeStep;
	real propagation;
	
	// config energy
 public:
	enum Properties { x = 0, y };
	static const int nLocalProps = 2;
	static const int nGlobalProps = 0;	
	
 private:
	// data
	int sParticles; /// allocated size of the particles array
	int nParticles; /// size of the particles array
	typedef real Particle[nLocalProps];
	Particle *localProps_[2];
	Particle *localProps, *localProps_new;
	real globalProps[nGlobalProps]; /// the common properties to all particles


 public:
	EdgeMP(real minDist, real maxDist, real stopThres, real timeStep);
	void setProperties(real minDist, real maxDist, real stopThres, real timeStep);
	void initialize(cyril::Image<uchar> &img, real center[nLocalProps], real radius, int propagation, real globalProps[nGlobalProps]);
	int converge(cyril::Image<uchar> &img);
	
 protected:
	void saveResult(cyril::Image<uchar> &img, cyril::Image<uchar> &img_disp, char *prefix, int num, char *suffix);
	real update(cyril::Image<uchar> &img);
	void reorganize();
	void resizeParticles(int nPart = 0);

};







} //snakes



#endif



